import 'dart:math' as math;
import 'package:flutter/material.dart';
import 'package:flutter_font_icons/flutter_font_icons.dart';
import 'package:flutter_smart_dialog/flutter_smart_dialog.dart';
import 'package:imsdk_client_core/imsdk_client_core.dart';

import 'app_config.dart';
import 'app_locale.dart';
import 'app_opener.dart';
import 'app_router.dart';
import 'session_amap.dart';
import 'session_message.dart';
import 'session_realtime.dart';
import 'widget/widget_emojis.dart';
import 'widget/widget_picture.dart';

///
///聊天页面
///
class SessionChat extends StatefulWidget {
  /* ******************************** instance content ******************************** */
  ///当前去会话
  final Session session;

  const SessionChat._({Key? key, required this.session}) : super(key: key);

  @override
  State<StatefulWidget> createState() => _SessionChatState();

  /* ******************************** static content ******************************** */
  ///当前SessionChat路由实例
  static MaterialPageRoute? _routeInstance;

  ///打开新的SessionChat路由实例，如果存在旧SessionChat路由实例则先移除
  static void pushRouteSingleton(BuildContext context, {required Session session, bool pushAndRemoveUntilFirst = false}) {
    _removeRouteInstance(context); //先移除可能存在的旧实例
    _routeInstance = MaterialPageRoute(builder: (context) => SessionChat._(session: session));
    if (pushAndRemoveUntilFirst) {
      Navigator.pushAndRemoveUntil(context, _routeInstance!, (route) => route.isFirst);
    } else {
      Navigator.push(context, _routeInstance!);
    }
  }

  ///直接从路由栈中移除SessionChat路由实例
  static void _removeRouteInstance(BuildContext context) {
    if (_routeInstance != null && _routeInstance!.isActive) {
      Navigator.removeRoute(context, _routeInstance!);
      _routeInstance = null;
    }
  }
}

enum _SessionChatKeyboardType { none, input, emoji, menu }

class _SessionChatState extends State<SessionChat> {
  static const _heroSessionChatMsgImgTag = '_heroSessionChatMsgImgTag';

  static const _inputBarIconButtonSize = 28.0;
  static const _keyboardIconButtonSize = 32.0;

  final _scrollController = ScrollController();
  final _valueController = TextEditingController(text: '');
  final _valueFocusNode = FocusNode();
  _SessionChatKeyboardType _keyboardType = _SessionChatKeyboardType.none;
  double _keyboardHeight = AppConfig.keyboardHeight;
  bool _isValueEmpty = true;
  bool _messageLoadOver = false; //是否已加载完
  bool _messageLoadLock = false; //是否正在加载

  @override
  void initState() {
    super.initState();
    AppConfig.logDebug(['SessionChat.initState()']);
    AppConfig.client.setOnUserFetchedWatcher(_clientCommonListener, remove: false);
    AppConfig.client.setOnTeamFetchedWatcher(_clientCommonListener, remove: false);
    AppConfig.client.setOnUserUpdateListener(_clientCommonListener, remove: false);
    AppConfig.client.setOnTeamUpdateListener(_clientCommonListener, remove: false);
    AppConfig.client.setOnUserShipUpdateListener(_clientCommonListener, remove: false);
    AppConfig.client.setOnTeamShipUpdateListener(_clientCommonListener, remove: false);
    AppConfig.client.setOnMessageUpdateListener(_clientCommonListener, remove: false);
    AppConfig.client.setOnMessageSendListener(_onMessageSendListener, remove: false);
    _loadMessageList(true);
  }

  @override
  void dispose() {
    AppConfig.logDebug(['SessionChat.dispose()']);
    _scrollController.dispose();
    _valueController.dispose();
    _valueFocusNode.dispose();
    AppConfig.client.setOnUserFetchedWatcher(_clientCommonListener, remove: true);
    AppConfig.client.setOnTeamFetchedWatcher(_clientCommonListener, remove: true);
    AppConfig.client.setOnUserUpdateListener(_clientCommonListener, remove: true);
    AppConfig.client.setOnTeamUpdateListener(_clientCommonListener, remove: true);
    AppConfig.client.setOnUserShipUpdateListener(_clientCommonListener, remove: true);
    AppConfig.client.setOnTeamShipUpdateListener(_clientCommonListener, remove: true);
    AppConfig.client.setOnMessageUpdateListener(_clientCommonListener, remove: true);
    AppConfig.client.setOnMessageSendListener(_onMessageSendListener, remove: true);
    _markMessageReaded(); //标记已读（调用后服务器会推送聊天对象的最新信息）
    AppConfig.setKeyboardHeight(_keyboardHeight); //保存键盘高度
    super.dispose();
  }

  @override
  Widget build(BuildContext context) {
    final themeData = Theme.of(context);
    final queryData = MediaQuery.of(context);
    final backgroundColor = themeData.brightness == Brightness.light ? Colors.grey[200] : null; //增强light状态white气泡的对比度
    final dividerColor = themeData.brightness == Brightness.light ? Colors.grey[350] : null; //分割线颜色
    _keyboardHeight = math.max(_keyboardHeight, queryData.viewInsets.bottom); //更新保存键盘高度
    return Scaffold(
      resizeToAvoidBottomInset: false,
      backgroundColor: backgroundColor,
      appBar: _buildAppBar(backgroundColor),
      body: _buildBody(themeData, queryData, dividerColor),
    );
  }

  AppBar _buildAppBar(Color? backgroundColor) {
    return AppBar(
      backgroundColor: backgroundColor,
      title: Text(widget.session.chatTitle),
      actions: [
        IconButton(
          icon: const Icon(Icons.more_horiz_outlined),
          tooltip: AppLocalizations.sessionChatActionBtnDetailTip,
          onPressed: _onDetailBtnPressed,
        )
      ],
    );
  }

  Widget _buildBody(ThemeData themeData, MediaQueryData queryData, Color? dividerColor) {
    return Column(
      children: [
        Divider(color: dividerColor),
        _buildBodyMessages(themeData),
        Divider(color: dividerColor),
        _buildBodyInputBar(themeData),
        _keyboardType == _SessionChatKeyboardType.none ? const SizedBox() : Divider(color: dividerColor),
        _buildBodyKeyboard(themeData, queryData, dividerColor),
      ],
    );
  }

  Widget _buildBodyMessages(ThemeData themeData) {
    return Expanded(
      child: GestureDetector(
        behavior: HitTestBehavior.opaque,
        child: ListView.builder(
          reverse: true,
          physics: AppConfig.bodyGlobalScrollPhysics,
          padding: const EdgeInsets.symmetric(vertical: SessionMessage.outerPaddingDefault),
          controller: _scrollController,
          itemCount: widget.session.msgcache.length + 1,
          itemBuilder: (context, index) {
            //加载项
            if (index == widget.session.msgcache.length) {
              if (_messageLoadOver) return const SizedBox();
              _loadMessageList(false); //尝试加载下一页
              return const Center(child: AppConfig.bodyGlobalPageLoading);
            }
            //消息项
            final item = widget.session.msgcache[index];
            final prevIndex = index + 1;
            Message? prevItem;
            if (prevIndex >= 0 && prevIndex < widget.session.msgcache.length) {
              prevItem = widget.session.msgcache[prevIndex];
            }
            return SessionMessage(
              message: item,
              prevMsg: prevItem,
              showNick: item.from == Constant.msgFromTeam,
              onGetPictures: _getMessagePictures,
              onGetImgHeroTag: _getMessageImgHeroTag,
            );
          },
        ),
        onPanDown: (a) => _setKeyboardType(_SessionChatKeyboardType.none), //使用onPanDown，使得比消息item子项先触发
      ),
    );
  }

  Widget _buildBodyInputBar(ThemeData themeData) {
    return Ink(
      padding: EdgeInsets.symmetric(vertical: AppConfig.bodyGlobalPadding.top / 2),
      child: Column(
        children: [
          Row(
            children: [
              IconButton(
                padding: EdgeInsets.zero,
                icon: Icon(_keyboardType == _SessionChatKeyboardType.emoji ? SimpleLineIcons.arrow_up_circle : SimpleLineIcons.emotsmile, size: _inputBarIconButtonSize),
                onPressed: () => _setKeyboardType(_keyboardType == _SessionChatKeyboardType.emoji ? _SessionChatKeyboardType.input : _SessionChatKeyboardType.emoji),
              ),
              Expanded(
                child: TextFormField(
                  controller: _valueController,
                  focusNode: _valueFocusNode,
                  scrollPhysics: AppConfig.bodyGlobalScrollPhysics,
                  keyboardType: TextInputType.text,
                  readOnly: _keyboardType != _SessionChatKeyboardType.input,
                  showCursor: _keyboardType == _SessionChatKeyboardType.input || _keyboardType == _SessionChatKeyboardType.emoji,
                  minLines: 1,
                  maxLines: 6,
                  decoration: InputDecoration(
                    filled: true,
                    fillColor: themeData.cardColor,
                    enabledBorder: themeData.inputDecorationTheme.border?.copyWith(borderSide: BorderSide(color: themeData.cardColor)),
                  ),
                  onChanged: _onTextFieldChange,
                  onTap: () => _setKeyboardType(_SessionChatKeyboardType.input),
                ),
              ),
              IconButton(
                padding: _isValueEmpty ? EdgeInsets.zero : const EdgeInsets.only(top: 3),
                icon: Icon(
                  _isValueEmpty ? (_keyboardType == _SessionChatKeyboardType.menu ? SimpleLineIcons.arrow_up_circle : SimpleLineIcons.plus) : MaterialCommunityIcons.send_circle,
                  size: _isValueEmpty ? _inputBarIconButtonSize : _inputBarIconButtonSize + 7,
                ),
                color: _isValueEmpty ? null : themeData.colorScheme.secondary,
                onPressed: _isValueEmpty ? () => _setKeyboardType(_keyboardType == _SessionChatKeyboardType.menu ? _SessionChatKeyboardType.input : _SessionChatKeyboardType.menu) : _onSendTextBtnPressed,
              ),
            ],
          ),
        ],
      ),
    );
  }

  Widget _buildBodyKeyboard(ThemeData themeData, MediaQueryData queryData, Color? dividerColor) {
    switch (_keyboardType) {
      case _SessionChatKeyboardType.none:
        return SizedBox(height: queryData.padding.bottom);
      case _SessionChatKeyboardType.input:
        return Container(
          height: _keyboardHeight,
          padding: EdgeInsets.only(bottom: queryData.padding.bottom),
          alignment: Alignment.center,
          child: Text(AppLocalizations.sessionChatKeyboardInputHint),
        );
      case _SessionChatKeyboardType.emoji:
        return Container(
          height: _keyboardHeight,
          padding: EdgeInsets.only(bottom: queryData.padding.bottom),
          child: WidgetEmojis(
            typeActiveColor: themeData.colorScheme.secondary,
            typeNormalColor: Colors.grey,
            dividerColor: dividerColor,
            pagePhysics: AppConfig.bodyGlobalScrollPhysics,
            gridPhysics: AppConfig.bodyGlobalScrollPhysics,
            typePhysics: AppConfig.bodyGlobalScrollPhysics,
            onEmojiSelected: _onKeyboardEmojiSelected,
            onBackspaceTap: _onKeyboardEmojiBackspaceTap,
          ),
        );
      case _SessionChatKeyboardType.menu:
        final menuList = [
          MaterialButton(
            padding: EdgeInsets.zero,
            onPressed: _onKeyboardMenuGalleryPressed,
            child: Column(mainAxisAlignment: MainAxisAlignment.center, children: [const Icon(Icons.image_outlined, size: _keyboardIconButtonSize), Text(AppLocalizations.sessionChatKeyboardMenuItemGallery, style: themeData.textTheme.caption)]),
          ),
          MaterialButton(
            padding: EdgeInsets.zero,
            onPressed: _onKeyboardMenuCameraPressed,
            child: Column(mainAxisAlignment: MainAxisAlignment.center, children: [const Icon(Icons.photo_camera_outlined, size: _keyboardIconButtonSize), Text(AppLocalizations.sessionChatKeyboardMenuItemCamera, style: themeData.textTheme.caption)]),
          ),
          MaterialButton(
            padding: EdgeInsets.zero,
            onPressed: _onKeyboardMenuRealVoicePressed,
            child: Column(mainAxisAlignment: MainAxisAlignment.center, children: [const Icon(Icons.phone_in_talk_outlined, size: _keyboardIconButtonSize), Text(AppLocalizations.sessionChatKeyboardMenuItemVoice, style: themeData.textTheme.caption)]),
          ),
          MaterialButton(
            padding: EdgeInsets.zero,
            onPressed: _onKeyboardMenuRealVideoPressed,
            child: Column(mainAxisAlignment: MainAxisAlignment.center, children: [const Icon(Icons.ondemand_video_outlined, size: _keyboardIconButtonSize), Text(AppLocalizations.sessionChatKeyboardMenuItemVideo, style: themeData.textTheme.caption)]),
          ),
        ];
        menuList.add(MaterialButton(
          padding: EdgeInsets.zero,
          onPressed: _onKeyboardMenuRealSharePressed,
          child: Column(mainAxisAlignment: MainAxisAlignment.center, children: [const Icon(Icons.airplay_outlined, size: _keyboardIconButtonSize), Text(AppLocalizations.sessionChatKeyboardMenuItemScreen, style: themeData.textTheme.caption)]),
        ));
        if (AppOpener.isAmapSupported()) {
          menuList.add(MaterialButton(
            padding: EdgeInsets.zero,
            onPressed: _onKeyboardMenuLocationPressed,
            child: Column(mainAxisAlignment: MainAxisAlignment.center, children: [const Icon(Icons.place_outlined, size: _keyboardIconButtonSize), Text(AppLocalizations.sessionChatKeyboardMenuItemLocaltion, style: themeData.textTheme.caption)]),
          ));
          menuList.add(MaterialButton(
            padding: EdgeInsets.zero,
            onPressed: _onKeyboardMenuRealLocalPressed,
            child: Column(mainAxisAlignment: MainAxisAlignment.center, children: [const Icon(Icons.my_location_outlined, size: _keyboardIconButtonSize), Text(AppLocalizations.sessionChatKeyboardMenuItemLocalreal, style: themeData.textTheme.caption)]),
          ));
        }
        menuList.add(MaterialButton(
          padding: EdgeInsets.zero,
          onPressed: _onKeyboardMenuRedpackPressed,
          child: Column(mainAxisAlignment: MainAxisAlignment.center, children: [const Icon(Icons.card_giftcard, size: _keyboardIconButtonSize), Text(AppLocalizations.sessionChatKeyboardMenuItemRedpack, style: themeData.textTheme.caption)]),
        ));
        return Container(
          height: _keyboardHeight,
          padding: EdgeInsets.only(bottom: queryData.padding.bottom),
          child: GridView.count(
            scrollDirection: Axis.vertical,
            physics: AppConfig.bodyGlobalScrollPhysics,
            crossAxisCount: 4,
            childAspectRatio: 1.2,
            children: menuList,
          ),
        );
    }
  }

  int _getMessagePictures(Message msg, List<String> urlList, List<String> heroList) {
    int index = 0;
    urlList.clear();
    for (var item in widget.session.msgcache) {
      if (item.type == Constant.msgTypeImage) {
        if (item.id == msg.id) index = urlList.length;
        urlList.add(item.shareLinkUrl);
        heroList.add(_getMessageImgHeroTag(item));
      }
    }
    return index;
  }

  String _getMessageImgHeroTag(Message msg) {
    return '$_heroSessionChatMsgImgTag${msg.id}';
  }

  void _setKeyboardType(_SessionChatKeyboardType type) {
    if (type != _keyboardType) {
      AppConfig.logDebug(['_setKeyboardType', type]);
      if (type == _SessionChatKeyboardType.input || type == _SessionChatKeyboardType.emoji) {
        _valueFocusNode.requestFocus();
      } else {
        _valueFocusNode.unfocus();
      }
      if (mounted) {
        setState(() {
          _keyboardType = type;
        });
      }
    }
  }

  void _onKeyboardEmojiSelected(String emoji) {
    AppConfig.insertCustomKeyboardText(_valueFocusNode, _valueController, emoji);
    _onTextFieldChange(_valueController.text); //手动触发onchange事件
  }

  void _onKeyboardEmojiBackspaceTap() {
    AppConfig.deleteCustomKeyboardText(_valueFocusNode, _valueController);
    _onTextFieldChange(_valueController.text); //手动触发onchange事件
  }

  void _onKeyboardMenuGalleryPressed() {
    WidgetPicture.pickFromGallery(context).then((value) async {
      if (value == null) return;
      final bytes = await AppRouter.pushImageEditorPage(context, type: WidgetPictureType.editAsJpg, editTitle: AppLocalizations.sessionChatImageEditTitle, editBytes: value, cropOnly: false, outputResize: false);
      if (bytes == null) return;
      SmartDialog.showLoading(msg: AppLocalizations.sessionChatFileUploadLoading);
      final result1 = await AppConfig.client.attachUpload(type: Constant.metaTypeMessage, fileBytes: [bytes], mediaType: MediaType.parse(WidgetPicture.mediaTypeJpg));
      SmartDialog.dismiss();
      if (result1.ok) {
        final result2 = await AppConfig.client.messageSendImage(sid: widget.session.sid, from: widget.session.msgfrom, shareLinkUrl: result1.extra!.first.path);
        if (!result2.ok) {
          SmartDialog.showToast(result2.desc);
        }
      } else {
        SmartDialog.showToast(result1.desc);
      }
    });
  }

  void _onKeyboardMenuCameraPressed() {
    WidgetPicture.pickFromCamera(context).then((value) async {
      if (value == null) return;
      final bytes = await AppRouter.pushImageEditorPage(context, type: WidgetPictureType.editAsJpg, editTitle: AppLocalizations.sessionChatImageEditTitle, editBytes: value, cropOnly: false, outputResize: false);
      if (bytes == null) return;
      SmartDialog.showLoading(msg: AppLocalizations.sessionChatFileUploadLoading);
      final result1 = await AppConfig.client.attachUpload(type: Constant.metaTypeMessage, fileBytes: [bytes], mediaType: MediaType.parse(WidgetPicture.mediaTypeJpg));
      SmartDialog.dismiss();
      if (result1.ok) {
        final result2 = await AppConfig.client.messageSendImage(sid: widget.session.sid, from: widget.session.msgfrom, shareLinkUrl: result1.extra!.first.path);
        if (!result2.ok) {
          SmartDialog.showToast(result2.desc);
        }
      } else {
        SmartDialog.showToast(result1.desc);
      }
    });
  }

  void _onKeyboardMenuRealVoicePressed() async {
    if (!SessionRealTime.canStartRealtimeChat()) {
      SmartDialog.showToast(AppLocalizations.sessionChatRealtimeDeviceIsBusy);
      return;
    }
    switch (widget.session.msgfrom) {
      case Constant.msgFromUser:
        final result = await AppConfig.client.messageSendRealtimeVoice(sid: widget.session.sid, from: widget.session.msgfrom, mediaUids: [AppConfig.client.user.id, widget.session.rid]);
        if (!result.ok) {
          SmartDialog.showToast(result.desc);
        }
        break;
      case Constant.msgFromTeam:
        AppRouter.pushContactTeamUsersPage(context, title: AppLocalizations.sessionChatTeamusersPickerTitle, team: widget.session.team, picker: true).then((value) async {
          if (value == null) return;
          final result = await AppConfig.client.messageSendRealtimeVoice(sid: widget.session.sid, from: widget.session.msgfrom, mediaUids: [AppConfig.client.user.id, ...value.map((e) => e.uid)]);
          if (!result.ok) {
            SmartDialog.showToast(result.desc);
          }
        });
        break;
    }
  }

  void _onKeyboardMenuRealVideoPressed() async {
    if (!SessionRealTime.canStartRealtimeChat()) {
      SmartDialog.showToast(AppLocalizations.sessionChatRealtimeDeviceIsBusy);
      return;
    }
    switch (widget.session.msgfrom) {
      case Constant.msgFromUser:
        final result = await AppConfig.client.messageSendRealtimeVideo(sid: widget.session.sid, from: widget.session.msgfrom, mediaUids: [AppConfig.client.user.id, widget.session.rid]);
        if (!result.ok) {
          SmartDialog.showToast(result.desc);
        }
        break;
      case Constant.msgFromTeam:
        AppRouter.pushContactTeamUsersPage(context, title: AppLocalizations.sessionChatTeamusersPickerTitle, team: widget.session.team, picker: true).then((value) async {
          if (value == null) return;
          final result = await AppConfig.client.messageSendRealtimeVideo(sid: widget.session.sid, from: widget.session.msgfrom, mediaUids: [AppConfig.client.user.id, ...value.map((e) => e.uid)]);
          if (!result.ok) {
            SmartDialog.showToast(result.desc);
          }
        });
        break;
    }
  }

  void _onKeyboardMenuRealSharePressed() async {
    if (!SessionRealTime.canStartRealtimeChat()) {
      SmartDialog.showToast(AppLocalizations.sessionChatRealtimeDeviceIsBusy);
      return;
    }
    switch (widget.session.msgfrom) {
      case Constant.msgFromUser:
        final result = await AppConfig.client.messageSendRealtimeShare(sid: widget.session.sid, from: widget.session.msgfrom, mediaUids: [AppConfig.client.user.id, widget.session.rid]);
        if (!result.ok) {
          SmartDialog.showToast(result.desc);
        }
        break;
      case Constant.msgFromTeam:
        AppRouter.pushContactTeamUsersPage(context, title: AppLocalizations.sessionChatTeamusersPickerTitle, team: widget.session.team, picker: true).then((value) async {
          if (value == null) return;
          final result = await AppConfig.client.messageSendRealtimeShare(sid: widget.session.sid, from: widget.session.msgfrom, mediaUids: [AppConfig.client.user.id, ...value.map((e) => e.uid)]);
          if (!result.ok) {
            SmartDialog.showToast(result.desc);
          }
        });
        break;
    }
  }

  void _onKeyboardMenuRealLocalPressed() async {
    if (!SessionRealTime.canStartRealtimeChat()) {
      SmartDialog.showToast(AppLocalizations.sessionChatRealtimeDeviceIsBusy);
      return;
    }
    switch (widget.session.msgfrom) {
      case Constant.msgFromUser:
        final result = await AppConfig.client.messageSendRealtimeLocal(sid: widget.session.sid, from: widget.session.msgfrom, mediaUids: [AppConfig.client.user.id, widget.session.rid]);
        if (!result.ok) {
          SmartDialog.showToast(result.desc);
        }
        break;
      case Constant.msgFromTeam:
        AppRouter.pushContactTeamUsersPage(context, title: AppLocalizations.sessionChatTeamusersPickerTitle, team: widget.session.team, picker: true).then((value) async {
          if (value == null) return;
          final result = await AppConfig.client.messageSendRealtimeLocal(sid: widget.session.sid, from: widget.session.msgfrom, mediaUids: [AppConfig.client.user.id, ...value.map((e) => e.uid)]);
          if (!result.ok) {
            SmartDialog.showToast(result.desc);
          }
        });
        break;
    }
  }

  void _onKeyboardMenuLocationPressed() async {
    final location = await AppRouter.pushSessionAmapPage(context, type: SessionAmapType.locationShare);
    if (location == null) return;
    final result = await AppConfig.client.messageSendShareLocation(sid: widget.session.sid, from: widget.session.msgfrom, shareLocation: location);
    if (!result.ok) {
      SmartDialog.showToast(result.desc);
    }
  }

  void _onKeyboardMenuRedpackPressed() {
    AppRouter.pushSessionRedpackPage(context, sid: widget.session.sid);
  }

  void _onTextFieldChange(String value) {
    final currValueEmpty = value.trim() == '';
    if (_isValueEmpty != currValueEmpty) {
      if (mounted) {
        setState(() {
          _isValueEmpty = currValueEmpty;
        });
      }
    }
  }

  void _onSendTextBtnPressed() {
    AppConfig.client.messageSendText(sid: widget.session.sid, from: widget.session.msgfrom, body: _valueController.text.trim()).then((value) {
      if (!value.ok) {
        SmartDialog.showToast(value.desc);
      }
    });
    _valueController.clear();
    if (mounted) {
      setState(() {
        _isValueEmpty = true;
      });
    }
  }

  void _onMessageSendListener(EasyPacket packet) {
    if (AppConfig.client.isMessageSendForSession(widget.session, packet)) {
      if (_scrollController.hasClients) {
        _scrollController.jumpTo(0);
      }
      if (mounted) {
        setState(() {});
      }
    }
  }

  void _clientCommonListener(EasyPacket packet) {
    final localship = AppConfig.client.getUserShip(widget.session.rid) ?? AppConfig.client.getTeamShip(widget.session.rid);
    if (localship == null) {
      SessionChat._removeRouteInstance(context);
    } else {
      if (mounted) {
        setState(() {});
      }
    }
  }

  void _onDetailBtnPressed() {
    switch (widget.session.msgfrom) {
      case Constant.msgFromUser:
        AppRouter.pushContactUserInfoPage(context, uid: widget.session.rid, fid: widget.session.fid, from: widget.session.from);
        break;
      case Constant.msgFromTeam:
        AppRouter.pushContactTeamInfoPage(context, tid: widget.session.rid, fid: widget.session.fid, from: widget.session.from);
        break;
    }
  }

  void _markMessageReaded() {
    switch (widget.session.msgfrom) {
      case Constant.msgFromUser:
        final ship = UserShipDirty();
        ship.unread = 0;
        AppConfig.client.userShipUpdate(id: widget.session.id, fields: ship.data);
        break;
      case Constant.msgFromTeam:
        final ship = TeamShipDirty();
        ship.unread = 0;
        AppConfig.client.teamShipUpdate(id: widget.session.id, fields: ship.data);
        break;
    }
  }

  void _loadMessageList(bool reload) {
    if (_messageLoadLock) return;
    _messageLoadLock = true; //加锁
    AppConfig.client.messageLoad(session: widget.session, reload: reload).then((value) {
      _messageLoadLock = false; //解锁
      if (value.ok) {
        if (mounted) {
          setState(() {
            _messageLoadOver = value.extra!;
          });
        }
      }
    }).catchError((e) {
      _messageLoadLock = false; //解锁
    });
  }
}
