import 'dart:async';
import 'dart:math' as math;

import 'package:flutter/material.dart';
import 'package:flutter_webrtc/flutter_webrtc.dart';
import 'package:just_audio/just_audio.dart';
import 'package:imsdk_client_core/imsdk_client_core.dart';

import '../sound/sound.dart';
import 'app_config.dart';
import 'app_locale.dart';
import 'widget/widget_avator.dart';
import 'widget/widget_image.dart';

///
///语音视频聊天页
///
class SessionRealTime extends StatefulWidget {
  /* ******************************** instance content ******************************** */

  ///关联的消息
  final Message message;

  const SessionRealTime._({Key? key, required this.message}) : super(key: key);

  @override
  State<StatefulWidget> createState() => _SessionRealTimeState();

  /* ******************************** static content ******************************** */
  ///占线标识，用来保证同时只有一个语音视频聊天页在路由栈中
  static bool _isCalling = false;

  ///当前SessionRealTime路由实例
  static PageRouteBuilder? _routeInstance;

  ///打开新的SessionRealTime路由实例，如果存在旧SessionRealTime路由实例则先移除
  static void pushRouteSingleton(BuildContext context, {required Message message}) {
    if (_isCalling) return;
    _isCalling = true; //最先设置占线标识
    _removeRouteInstance(context); //先移除可能存在的旧实例
    _routeInstance = PageRouteBuilder(
      pageBuilder: (context, animation, secondaryAnimation) => SessionRealTime._(message: message),
      transitionsBuilder: (context, animation, secondaryAnimation, child) => ScaleTransition(scale: animation, child: FadeTransition(opacity: animation, child: child)),
    );
    Navigator.push(context, _routeInstance!);
  }

  ///直接从路由栈中移除SessionRealTime路由实例
  static void _removeRouteInstance(BuildContext context) {
    if (_routeInstance != null && _routeInstance!.isActive) {
      Navigator.removeRoute(context, _routeInstance!);
      _routeInstance = null;
    }
  }

  ///是否可以启用语音视频聊天组建
  static bool canStartRealtimeChat() => !_isCalling;
}

class _SessionRealTimeState extends State<SessionRealTime> {
  static const _iconButtonBgSize = 66.0;
  static const _iconButtonIcSize = 38.0;

  final _localIdent = DateTime.now().millisecondsSinceEpoch; //本地实例毫秒时间戳标志
  final _renders = <RTCVideoRenderer>[];
  final _streams = <MediaStream?>[];
  final _connections = <RTCPeerConnection?>[];
  final _initializeIdents = <int>[]; //毫秒时间戳初始化标志: <=0表示本地未初始化 或 未收到远程已初始化的信令，>0表示本地或远程初始化完毕的毫秒时间戳标志
  final _cachedCandidates = <List<RTCIceCandidate>>[]; //缓存的每个远程连接的ice信息

  AudioPlayer? _ringPlayer; //来电提示铃声播放器
  Timer? _heartbeatTimer; //循环发送初始化信令的定时器
  bool _micphoneIsEnable = true; //麦克风是否启用
  bool _speakerphoneIsEnable = true; //扬声器是否启用
  bool _frontcameraIsEnable = true; //前置摄像头是否启用
  Offset? _p2pSlaveMediaViewOffset; //P2P小媒体视图坐标

  String get sdpSemantics => WebRTC.platformIsWindows ? 'plan-b' : 'unified-plan';
  bool get hasVideoMedia => widget.message.type == Constant.msgTypeRealtimeVideo || widget.message.type == Constant.msgTypeRealtimeShare; //是否包含视频
  bool get isMessageSender => widget.message.uid == AppConfig.client.user.id; //本地用户是否为消息发起者
  bool get isP2PMessage => widget.message.from == Constant.msgFromUser; //是否为P2P消息
  void _setInitialized(int index, int ident) => _initializeIdents[index] = ident; //设置对应端初始化状态的毫秒时间戳标志
  bool _wasInitialized(int index) => _initializeIdents[index] > 0; //对应端是否已经初始化完毕
  bool _keepConnection(int remoteIndex, int remoteIdent) => _initializeIdents[remoteIndex] == remoteIdent; //对应端是否保持当前连接，即对应端连接不需要重置
  bool _hasAnyConnect() => _connections.any((element) => element != null); //是否有远程端连接已经建立
  int _getLocalIndex() => widget.message.mediaJoined.indexOf(AppConfig.client.user.id); //ObjectId已经重定义hashCode函数与==操作符，可以直接用indexOf
  bool _wasP2PInitialized() => _wasInitialized(0) && _wasInitialized(1); //P2P连接已初始化完毕
  RTCVideoRenderer _getP2PMasterRender() => _wasP2PInitialized() ? _renders[_getLocalIndex() == 0 ? 1 : 0] : _renders[_getLocalIndex()]; //P2P大视图窗口render
  RTCVideoRenderer _getP2PSlaverRender() => _wasP2PInitialized() ? _renders[_getLocalIndex() == 0 ? 0 : 1] : _renders[_getLocalIndex()]; //P2P小视图窗口render
  ObjectId _getP2PRemoteUid() => widget.message.mediaJoined[_getLocalIndex() == 0 ? 1 : 0]; //P2P远程用户id

  @override
  void initState() {
    super.initState();
    AppConfig.logDebug(['SessionRealTime.initState()', widget.message.mediaJoined]);
    for (int i = 0; i < widget.message.mediaJoined.length; i++) {
      _renders.add(RTCVideoRenderer());
      _streams.add(null);
      _connections.add(null);
      _initializeIdents.add(-1);
      _cachedCandidates.add([]);
    }
    AppConfig.client.setOnMessageUpdateListener(_onMessageUpdateListener, remove: false);
    AppConfig.client.setOnMessageWebrtcListener(_onMessageWebrtcListener, remove: false);
    _initDevices();
  }

  @override
  void dispose() {
    AppConfig.logDebug(['SessionRealTime.dispose()', widget.message.mediaJoined]);
    AppConfig.client.setOnMessageUpdateListener(_onMessageUpdateListener, remove: true);
    AppConfig.client.setOnMessageWebrtcListener(_onMessageWebrtcListener, remove: true);
    _closeDevices();
    super.dispose();
  }

  @override
  Widget build(BuildContext context) {
    final themeData = Theme.of(context);
    final queryData = MediaQuery.of(context);
    final backgroundColor = themeData.brightness == Brightness.light ? Colors.grey[300] : null; //增强light状态white气泡的对比度
    return WillPopScope(
      child: Scaffold(
        resizeToAvoidBottomInset: false,
        backgroundColor: backgroundColor,
        extendBodyBehindAppBar: true,
        appBar: AppBar(
          backgroundColor: Colors.transparent,
          systemOverlayStyle: hasVideoMedia && isP2PMessage ? AppConfig.darkSystemUiOverlayStyle : AppConfig.systemUiOverlayStyle,
          automaticallyImplyLeading: false,
          toolbarHeight: 0,
        ),
        body: Stack(
          children: [
            isP2PMessage ? _buildP2PMasterMediaView(themeData, queryData) : const SizedBox(),
            isP2PMessage ? _buildP2PSlaverMediaView(themeData, queryData) : const SizedBox(),
            isP2PMessage ? const SizedBox() : _buildGridMediaView(themeData, queryData),
            Align(
              alignment: Alignment.bottomCenter,
              child: Padding(
                padding: EdgeInsets.only(bottom: queryData.padding.bottom + AppConfig.bodyGlobalPadding.bottom * 3),
                child: Column(
                  mainAxisSize: MainAxisSize.min,
                  children: [
                    Text(
                      _hasAnyConnect() ? AppLocalizations.sessionRealtimeTextChating : AppLocalizations.sessionRealtimeTextWaiting,
                      style: themeData.textTheme.subtitle1?.copyWith(color: hasVideoMedia && isP2PMessage ? Colors.white : null),
                    ),
                    SizedBox(height: AppConfig.bodyGlobalPadding.bottom * 2),
                    Row(
                      mainAxisSize: MainAxisSize.max,
                      mainAxisAlignment: MainAxisAlignment.center,
                      children: [
                        _heartbeatTimer == null ? const SizedBox() : _buildLeftButton(themeData, queryData),
                        _heartbeatTimer == null ? const SizedBox() : const SizedBox(width: _iconButtonBgSize / 2),
                        _buildCenterButton(themeData, queryData),
                        const SizedBox(width: _iconButtonBgSize / 2),
                        _buildRightButton(themeData, queryData),
                      ],
                    )
                  ],
                ),
              ),
            ),
          ],
        ),
      ),
      onWillPop: () async => false, //拦截返回按钮
    );
  }

  Widget _buildP2PMasterMediaView(ThemeData themeData, MediaQueryData queryData) {
    if (hasVideoMedia) {
      return GestureDetector(
        behavior: HitTestBehavior.opaque,
        child: SizedBox(
          width: queryData.size.width,
          height: queryData.size.height,
          child: RTCVideoView(_getP2PMasterRender(), objectFit: RTCVideoViewObjectFit.RTCVideoViewObjectFitCover),
        ),
        onTap: () {},
      );
    } else {
      return const SizedBox();
    }
  }

  Widget _buildP2PSlaverMediaView(ThemeData themeData, MediaQueryData queryData) {
    if (hasVideoMedia && _wasP2PInitialized()) {
      final viewSize = Size(queryData.size.width / 4, queryData.size.height / 4);
      _p2pSlaveMediaViewOffset ??= Offset(queryData.size.width - viewSize.width - AppConfig.bodyGlobalPadding.right, queryData.padding.top + AppConfig.bodyGlobalPadding.top);
      return Positioned(
        left: _p2pSlaveMediaViewOffset!.dx,
        top: _p2pSlaveMediaViewOffset!.dy,
        child: GestureDetector(
          behavior: HitTestBehavior.opaque,
          child: SizedBox(
            width: viewSize.width,
            height: viewSize.height,
            child: RTCVideoView(_getP2PSlaverRender(), objectFit: RTCVideoViewObjectFit.RTCVideoViewObjectFitCover),
          ),
          onTap: () {},
          onPanUpdate: (details) {
            final offset = _calcP2PSlaverViewOffset(queryData.size, viewSize, _p2pSlaveMediaViewOffset!, details.delta);
            if (mounted) {
              setState(() {
                _p2pSlaveMediaViewOffset = offset;
              });
            }
          },
        ),
      );
    } else {
      final ship = AppConfig.client.getUserShip(_getP2PRemoteUid());
      final user = AppConfig.client.getUser(_getP2PRemoteUid());
      if (ship == null) {
        return const SizedBox();
      } else {
        return Align(
          alignment: Alignment.topCenter,
          child: Padding(
            padding: EdgeInsets.only(top: queryData.size.height / 4),
            child: Column(
              children: [
                WidgetAvator(type: WidgetAvatorType.huge, url: user.icon, urlList: user.head, onFormatUrl: AppConfig.formatAttachUrl),
                SizedBox(height: AppConfig.bodyGlobalPadding.top / 2),
                Text(
                  ComTools.formatCNcharacters(ComTools.formatUserShipNick(ship, user)),
                  style: themeData.textTheme.subtitle1?.copyWith(color: hasVideoMedia && isP2PMessage ? Colors.white : null),
                ),
              ],
            ),
          ),
        );
      }
    }
  }

  Offset _calcP2PSlaverViewOffset(Size boxSize, Size viewSize, Offset offset, Offset delta) {
    //水平方向偏移量不能小于0不能大于屏幕最大宽度
    double dx = 0;
    if (offset.dx + delta.dx < 0) {
      dx = 0;
    } else if (offset.dx + delta.dx > boxSize.width - viewSize.width) {
      dx = boxSize.width - viewSize.width;
    } else {
      dx = offset.dx + delta.dx;
    }
    //垂直方向偏移量不能小于0不能大于屏幕最大高度
    double dy = 0;
    if (offset.dy + delta.dy < 0) {
      dy = 0;
    } else if (offset.dy + delta.dy > boxSize.height - viewSize.height) {
      dy = boxSize.height - viewSize.height;
    } else {
      dy = offset.dy + delta.dy;
    }
    return Offset(dx, dy);
  }

  Widget _buildGridMediaView(ThemeData themeData, MediaQueryData queryData) {
    final crossAxisCount = widget.message.mediaJoined.length <= 4 ? 2 : 3;
    final itemSize = math.min(queryData.size.width, queryData.size.height) / crossAxisCount;
    return GridView.builder(
      scrollDirection: Axis.vertical,
      physics: AppConfig.bodyGlobalScrollPhysics,
      gridDelegate: SliverGridDelegateWithFixedCrossAxisCount(crossAxisCount: crossAxisCount),
      itemCount: widget.message.mediaJoined.length,
      itemBuilder: (context, gridIndex) {
        final user = AppConfig.client.getUser(widget.message.mediaJoined[gridIndex]);
        return Container(
          width: itemSize,
          height: itemSize,
          color: Colors.black,
          child: Stack(
            alignment: Alignment.center,
            children: [
              Positioned(top: 0, bottom: 0, right: 0, left: 0, child: WidgetImage(url: user.icon, onFormatUrl: AppConfig.formatAttachUrl)),
              Positioned(top: 0, bottom: 0, right: 0, left: 0, child: hasVideoMedia ? RTCVideoView(_renders[gridIndex], objectFit: RTCVideoViewObjectFit.RTCVideoViewObjectFitCover) : const SizedBox()),
            ],
          ),
        );
      },
    );
  }

  Widget _buildLeftButton(ThemeData themeData, MediaQueryData queryData) {
    //开关麦克风
    return GestureDetector(
      behavior: HitTestBehavior.opaque,
      child: Container(
        width: _iconButtonBgSize,
        height: _iconButtonBgSize,
        decoration: const ShapeDecoration(color: Colors.white, shape: RoundedRectangleBorder(borderRadius: BorderRadius.all(Radius.circular(_iconButtonBgSize)))),
        child: Icon(_micphoneIsEnable ? Icons.mic : Icons.mic_off, color: Colors.black, size: _iconButtonIcSize - 4),
      ),
      onTap: () {
        final localStream = _streams[_getLocalIndex()];
        if (localStream != null) {
          final audioTracks = localStream.getAudioTracks();
          if (audioTracks.isNotEmpty) {
            if (mounted) {
              setState(() {
                _micphoneIsEnable = !_micphoneIsEnable;
                audioTracks.first.enabled = _micphoneIsEnable; //开关麦克风
              });
            }
          }
        }
      },
    );
  }

  Widget _buildCenterButton(ThemeData themeData, MediaQueryData queryData) {
    //挂断电话
    return GestureDetector(
      behavior: HitTestBehavior.opaque,
      child: Container(
        width: _iconButtonBgSize,
        height: _iconButtonBgSize,
        decoration: const ShapeDecoration(color: Colors.red, shape: RoundedRectangleBorder(borderRadius: BorderRadius.all(Radius.circular(_iconButtonBgSize)))),
        child: const Icon(Icons.call_end, color: Colors.white, size: _iconButtonIcSize),
      ),
      onTap: () {
        if (_wasInitialized(_getLocalIndex())) {
          //是P2P会话 或 本人是会话发起者，则立即通知服务器结束会话
          if (isP2PMessage || isMessageSender) {
            AppConfig.client.messageUpdate(id: widget.message.id, realtimeEnd: true);
          }
          SessionRealTime._removeRouteInstance(context);
        }
      },
    );
  }

  Widget _buildRightButton(ThemeData themeData, MediaQueryData queryData) {
    if (hasVideoMedia) {
      if (_heartbeatTimer == null) {
        //接听电话
        return GestureDetector(
          behavior: HitTestBehavior.opaque,
          child: Container(
            width: _iconButtonBgSize,
            height: _iconButtonBgSize,
            decoration: const ShapeDecoration(color: Colors.green, shape: RoundedRectangleBorder(borderRadius: BorderRadius.all(Radius.circular(_iconButtonBgSize)))),
            child: const Icon(Icons.videocam, color: Colors.white, size: _iconButtonIcSize - 8),
          ),
          onTap: () {
            if (mounted) {
              setState(() {
                _resetHeartbeatTimer(byTap: true);
              });
            }
          },
        );
      } else {
        //切换摄像头
        return GestureDetector(
          behavior: HitTestBehavior.opaque,
          child: Container(
            width: _iconButtonBgSize,
            height: _iconButtonBgSize,
            decoration: const ShapeDecoration(color: Colors.white, shape: RoundedRectangleBorder(borderRadius: BorderRadius.all(Radius.circular(_iconButtonBgSize)))),
            child: const Icon(Icons.flip_camera_ios, color: Colors.black, size: _iconButtonIcSize - 8),
          ),
          onTap: () {
            final localStream = _streams[_getLocalIndex()];
            if (localStream != null) {
              final videoTracks = localStream.getVideoTracks();
              if (videoTracks.isNotEmpty) {
                Helper.switchCamera(videoTracks.first).then((value) {
                  if (mounted) {
                    setState(() {
                      _frontcameraIsEnable = value; //value为true时开启的是前置摄像头。value为false时开启的是后置摄像头。
                    });
                  }
                });
              }
            }
          },
        );
      }
    } else {
      if (_heartbeatTimer == null) {
        //接听电话
        return GestureDetector(
          behavior: HitTestBehavior.opaque,
          child: Container(
            width: _iconButtonBgSize,
            height: _iconButtonBgSize,
            decoration: const ShapeDecoration(color: Colors.green, shape: RoundedRectangleBorder(borderRadius: BorderRadius.all(Radius.circular(_iconButtonBgSize)))),
            child: const Icon(Icons.phone, color: Colors.white, size: _iconButtonIcSize),
          ),
          onTap: () {
            if (mounted) {
              setState(() {
                _resetHeartbeatTimer(byTap: true);
              });
            }
          },
        );
      } else {
        //开关扬声器
        return GestureDetector(
          behavior: HitTestBehavior.opaque,
          child: Container(
            width: _iconButtonBgSize,
            height: _iconButtonBgSize,
            decoration: const ShapeDecoration(color: Colors.white, shape: RoundedRectangleBorder(borderRadius: BorderRadius.all(Radius.circular(_iconButtonBgSize)))),
            child: Icon(_speakerphoneIsEnable ? Icons.volume_up : Icons.volume_off, color: Colors.black, size: _iconButtonIcSize - 8),
          ),
          onTap: () {
            final localStream = _streams[_getLocalIndex()];
            if (localStream != null) {
              final audioTracks = localStream.getAudioTracks();
              if (audioTracks.isNotEmpty) {
                if (mounted) {
                  setState(() {
                    _speakerphoneIsEnable = !_speakerphoneIsEnable;
                    if (AppConfig.isMobileDevice) audioTracks.first.enableSpeakerphone(_speakerphoneIsEnable); //开关扬声器
                  });
                }
              }
            }
          },
        );
      }
    }
  }

  void _initDevices() async {
    //初始化_renders
    for (var render in _renders) {
      await render.initialize();
    }
    //初始化本地设备stream
    final localIndex = _getLocalIndex();
    final localStream = _streams[localIndex] = await navigator.mediaDevices.getUserMedia({
      //经过测试audio不加去噪音参数时，两台手机靠近声音循环到一定程度时ios设备闪退，加上这下面两个参数会好很多
      //audio约束参考： https://blog.addpipe.com/audio-constraints-getusermedia/
      'audio': {
        'noiseSuppression': true,
        'echoCancellation': true,
      },
      'video': hasVideoMedia
          ? {
              'mandatory': {'minWidth': '320', 'minHeight': '320', 'minFrameRate': '30'},
              'facingMode': _frontcameraIsEnable ? 'user' : 'environment', //初始化设置摄像头状态，user对应前置摄像头
            }
          : false,
    });
    final audioTracks = localStream.getAudioTracks();
    if (audioTracks.isNotEmpty) {
      if (AppConfig.isMobileDevice) audioTracks.first.enableSpeakerphone(_speakerphoneIsEnable); //初始化设置扬声器状态
      audioTracks.first.enabled = _micphoneIsEnable; //初始化设置麦克风状态
    }
    //初始化来电提示铃声播放器，经测试在本地设备stream之后调用可避免很多问题
    _ringPlayer = AudioPlayer();
    await _ringPlayer?.setAsset(isMessageSender ? SoundAssets.realtimeRingCall : SoundAssets.realtimeRingWait);
    await _ringPlayer?.setLoopMode(LoopMode.all);
    _ringPlayer?.play();
    //由于该函数是在initState()中调用的，所以刷新界面需要进行延迟
    Future.delayed(Duration.zero, () {
      if (mounted) {
        setState(() {
          _renders[localIndex].srcObject = localStream;
          _setInitialized(localIndex, _localIdent); //先标记本地媒体已初始化完毕
          if (isMessageSender) _resetHeartbeatTimer(byTap: false); //消息发起者直接初始化心跳定时器
        });
      }
    });
  }

  void _closeDevices() async {
    //release _heartbeatTimer
    _heartbeatTimer?.cancel();
    _heartbeatTimer = null;
    AppConfig.logInfo(['_closeDevices release _heartbeatTimer success.']);
    //release _ringPlayer
    try {
      await _ringPlayer?.dispose();
      _ringPlayer = null;
      AppConfig.logInfo(['_closeDevices release _ringPlayer success.']);
    } catch (error, stack) {
      AppConfig.logError(['_closeDevices release _ringPlayer =>', error, '\n', stack]);
    }
    //release _streams
    try {
      for (var stream in _streams) {
        if (stream != null) {
          final tracks = stream.getTracks();
          for (var element in tracks) {
            await element.stop();
          }
          await stream.dispose();
        }
      }
      AppConfig.logInfo(['_closeDevices release _streams success.']);
    } catch (error, stack) {
      AppConfig.logError(['_closeDevices release _streams =>', error, '\n', stack]);
    }
    //release _connections
    try {
      for (var connection in _connections) {
        if (connection != null) {
          await connection.close();
        }
      }
      AppConfig.logInfo(['_closeDevices release _connections success.']);
    } catch (error, stack) {
      AppConfig.logError(['_closeDevices release _connections =>', error, '\n', stack]);
    }
    //release _renders
    try {
      for (var render in _renders) {
        await render.dispose();
      }
      AppConfig.logInfo(['_closeDevices release _renders success.']);
    } catch (error, stack) {
      AppConfig.logError(['_closeDevices release _renders =>', error, '\n', stack]);
    }
    //最后释放占线标志
    SessionRealTime._isCalling = false;
  }

  void _resetHeartbeatTimer({required bool byTap}) {
    //释放旧的
    _heartbeatTimer?.cancel();
    _heartbeatTimer = null;
    //创建新的
    if (_wasInitialized(_getLocalIndex())) {
      _heartbeatTimer = Timer.periodic(const Duration(seconds: Validator.webrtcInterval), (timer) {
        _sendReady();
      });
      _sendReady();
      //是P2P会话且是通过点击调用的 或 不是P2P会话且不是通过点击调用的，则立即通知服务器开始会话
      if ((isP2PMessage && byTap) || (!isP2PMessage && !byTap)) {
        AppConfig.client.messageUpdate(id: widget.message.id, realtimeStart: true);
      }
    }
  }

  Future<RTCPeerConnection?> _resetConnection(int localIndex, int remoteIndex, int remoteIdent) async {
    final localStream = _streams[localIndex];
    if (localStream == null) return null;
    //释放旧的
    try {
      await _ringPlayer?.dispose();
      _ringPlayer = null;
    } catch (error, stack) {
      AppConfig.logError(['_resetConnection release _ringPlayer =>', error, '\n', stack]);
    }
    try {
      final stream = _streams[remoteIndex];
      if (stream != null) {
        final tracks = stream.getTracks();
        for (var element in tracks) {
          await element.stop();
        }
        await stream.dispose();
      }
    } catch (error, stack) {
      AppConfig.logError(['_resetConnection release stream =>', error, '\n', stack]);
    }
    try {
      final connection = _connections[remoteIndex];
      if (connection != null) {
        await connection.close();
      }
    } catch (error, stack) {
      AppConfig.logError(['_resetConnection release connection =>', error, '\n', stack]);
    }
    //重置参数
    if (mounted) {
      setState(() {
        _renders[remoteIndex].srcObject = null;
        _streams[remoteIndex] = null;
        _connections[remoteIndex] = null;
        _setInitialized(remoteIndex, -1);
        _cachedCandidates[remoteIndex].clear();
      });
    }
    //创建连接
    final connection = _connections[remoteIndex] = await createPeerConnection({
      'iceServers': [
        {'url': 'stun:stun.l.google.com:19302'},
      ],
      'sdpSemantics': sdpSemantics,
    }, {
      'mandatory': {},
      'optional': [
        {'DtlsSrtpKeyAgreement': true},
      ]
    });
    //监听连接ice信息
    connection.onIceCandidate = (candidate) {
      _sendCandidate(remoteIndex, remoteIdent, candidate);
    };
    //监听连接远程流开启
    switch (sdpSemantics) {
      case 'plan-b':
        connection.onAddStream = (MediaStream stream) {
          if (hasVideoMedia) {
            final remoteStream = _streams[remoteIndex] = stream;
            if (mounted) {
              setState(() {
                _renders[remoteIndex].srcObject = remoteStream;
              });
            }
          }
        };
        await connection.addStream(localStream);
        break;
      case 'unified-plan':
        connection.onTrack = (event) {
          if (hasVideoMedia && event.track.kind == 'video') {
            final remoteStream = _streams[remoteIndex] = event.streams[0];
            if (mounted) {
              setState(() {
                _renders[remoteIndex].srcObject = remoteStream;
              });
            }
          }
        };
        final localTracks = localStream.getTracks();
        for (var track in localTracks) {
          await connection.addTrack(track, localStream);
        }
        break;
    }
    //标记远程连接初始化完毕
    if (mounted) {
      setState(() {
        _setInitialized(remoteIndex, remoteIdent);
      });
    }
    return connection;
  }

  void _sendReady() {
    final localIndex = _getLocalIndex(); //自己的索引
    AppConfig.client.messageWebrtc(id: widget.message.id, toUid: null, signals: {
      'sid': widget.message.sid,
      'type': 'ready',
      'ident': _localIdent,
      'index': localIndex,
      'data': {},
    });
  }

  void _sendCandidate(int remoteIndex, int remoteIdent, RTCIceCandidate candidate) {
    final localIndex = _getLocalIndex(); //自己的索引
    final toUid = widget.message.mediaJoined[remoteIndex]; //信令发送目标
    AppConfig.client.messageWebrtc(id: widget.message.id, toUid: toUid, signals: {
      'sid': widget.message.sid,
      'type': 'candidate',
      'ident': _localIdent,
      'index': localIndex,
      'data': {'targetIdent': remoteIdent, 'candidate': candidate.candidate, 'sdpMid': candidate.sdpMid, 'sdpMLineIndex': candidate.sdpMLineIndex},
    });
  }

  void _onMessageUpdateListener(EasyPacket packet) {
    final message = AppConfig.client.parseMessageFromSendOrUpdate(packet);
    if (message.id == widget.message.id && !message.mediaGoing) {
      SessionRealTime._removeRouteInstance(context);
    }
  }

  void _onMessageWebrtcListener(EasyPacket packet) async {
    final Map<String, dynamic> signals = packet.data!;
    final String sid = signals['sid']; //信令类型
    final String type = signals['type']; //信令类型
    final int remoteIdent = signals['ident']; //信令发送者的标志
    final int remoteIndex = signals['index']; //信令发送者的索引
    final Map<String, dynamic> data = signals['data']; //信令数据
    if (sid != widget.message.sid.toHexString()) return; //忽略会话id与当前消息不一致的
    final localIndex = _getLocalIndex(); //自己的索引
    if (localIndex == remoteIndex) return; //自己发送的
    final toUid = widget.message.mediaJoined[remoteIndex]; //信令发送目标
    if (!_wasInitialized(localIndex)) return; //本地未初始化完毕
    if (_heartbeatTimer == null) return; //本地已接听电话
    switch (type) {
      case 'ready':
        if (localIndex >= remoteIndex) return; //当本地索引>=远程索引时忽略这个预备信令
        if (_keepConnection(remoteIndex, remoteIdent)) return; //不需要重置连接则直接返回
        final connection = await _resetConnection(localIndex, remoteIndex, remoteIdent);
        if (connection == null) return;
        //发起offer
        final description = await connection.createOffer({});
        await connection.setLocalDescription(description);
        AppConfig.client.messageWebrtc(id: widget.message.id, toUid: toUid, signals: {
          'sid': widget.message.sid,
          'type': 'offer',
          'ident': _localIdent,
          'index': localIndex,
          'data': {'targetIdent': remoteIdent, 'sdp': description.sdp, 'type': description.type},
        });
        return;
      case 'offer':
        final int targetIdent = data['targetIdent'];
        if (targetIdent != _localIdent) return;
        final connection = await _resetConnection(localIndex, remoteIndex, remoteIdent);
        if (connection == null) return;
        //响应answer
        await connection.setRemoteDescription(RTCSessionDescription(data['sdp'], data['type']));
        final description = await connection.createAnswer({});
        await connection.setLocalDescription(description);
        AppConfig.client.messageWebrtc(id: widget.message.id, toUid: toUid, signals: {
          'sid': widget.message.sid,
          'type': 'answer',
          'ident': _localIdent,
          'index': localIndex,
          'data': {'targetIdent': remoteIdent, 'sdp': description.sdp, 'type': description.type},
        });
        //加入candidate
        final candidateList = _cachedCandidates[remoteIndex];
        if (candidateList.isNotEmpty) {
          for (var candidate in candidateList) {
            await connection.addCandidate(candidate);
          }
          candidateList.clear();
        }
        return;
      case 'answer':
        final int targetIdent = data['targetIdent'];
        if (targetIdent != _localIdent) return;
        final connection = _connections[remoteIndex];
        if (connection == null) return;
        //收到answer
        await connection.setRemoteDescription(RTCSessionDescription(data['sdp'], data['type']));
        //加入candidate
        final candidateList = _cachedCandidates[remoteIndex];
        if (candidateList.isNotEmpty) {
          for (var candidate in candidateList) {
            await connection.addCandidate(candidate);
          }
          candidateList.clear();
        }
        return;
      case 'candidate':
        final int targetIdent = data['targetIdent'];
        if (targetIdent != _localIdent) return;
        final connection = _connections[remoteIndex];
        final candidateList = _cachedCandidates[remoteIndex];
        //加入candidate
        final candidate = RTCIceCandidate(data['candidate'], data['sdpMid'], data['sdpMLineIndex']);
        if (connection == null || await connection.getRemoteDescription() == null) {
          candidateList.add(candidate);
        } else {
          await connection.addCandidate(candidate); //一定机率抛出异常，临时解决方案参考： https://github.com/flutter-webrtc/flutter-webrtc/issues/787#issuecomment-991483131
        }
        return;
    }
  }
}
